"""
Tests for the :class:`cpymad.madx.Madx` API.
"""

import os
import sys

import numpy as np
from numpy.testing import assert_allclose, assert_equal
from pytest import approx, fixture, mark, raises

import cpymad
from cpymad.madx import Madx, Sequence, metadata


@fixture
def mad():
    with Madx(prompt='X:> ') as mad:
        yield mad


@fixture
def lib(mad):
    return mad._libmadx


SEQU = """
! constants
QP_K1 = 2;

! elements
qp: quadrupole, k1:=QP_K1, l=1;
sb: sbend, l=2, angle=3.14/4;
dr: drift, l=1;

! sequences
s1: sequence, l=8, refer=center;
dr, at=0.5; ! dr[1] ~ betx_full1[1]
qp, at=1.5;
dr, at=2.5; ! dr[2] ~ betx_full1[3] ~ betx_range[0]
qp, at=3.5; !       ~ betx_full1[4] ~ betx_range[1]
dr, at=4.5;
sb, at=6.0; !                       ~ betx_range[3]
dr, at=7.5;
endsequence;

s2: sequence, l=3, refer=entry;
qp1: qp, at=0, k1=3;
qp2: qp, at=1, l=2;
endsequence;
"""


def normalize(path):
    """Normalize path name to eliminate different spellings of the same path.
    This is needed for path comparisons in tests, especially on windows where
    pathes are case insensitive and allow a multitude of spellings."""
    return os.path.normcase(os.path.normpath(path))


def test_copyright():
    notice = cpymad.get_copyright_notice()
    assert isinstance(notice, type(u""))


def test_version(mad):
    """Check that the Madx.version attribute can be used as expected."""
    version = mad.version
    # check format:
    major, minor, micro = map(int, version.release.split('.'))
    # We need at least MAD-X 5.05.00:
    assert (major, minor, micro) >= (5, 5, 0)
    # check format:
    year, month, day = map(int, version.date.split('.'))
    assert (year, month, day) >= (2019, 5, 10)
    assert 1 <= month <= 12
    assert 1 <= day <= 31
    assert str(version).startswith(
        'MAD-X {}'.format(version.release))


def test_metadata(mad):
    version = mad.version
    assert metadata.__version__ == version.release
    assert isinstance(metadata.get_copyright_notice(), type(u""))


def test_independent_instances():
    # Check independence by defining a variable differently in each
    # instance:
    with Madx(prompt='X1:> ') as mad1, Madx(prompt='X2:> ') as mad2:
        mad1.input('ANSWER=42;')
        mad2.input('ANSWER=43;')
        assert mad1.eval('ANSWER') == 42
        assert mad2.eval('ANSWER') == 43


# TODO: We need to fix this on windows, but for now, I just need it to
# pass so that the CI builds the release...
@mark.xfail(
    sys.platform != 'linux',
    reason='Output is sometimes garbled on MacOS and windows.',
)
def test_streamreader():
    output = []
    with Madx(stdout=output.append) as m:
        assert len(output) == 1
        assert b'+++++++++++++++++++++++++++++++++' in output[0]
        assert b'+ Support: mad@cern.ch,'           in output[0]
        assert b'+ Release   date: '                in output[0]
        assert b'+ Execution date: '                in output[0]
        # assert b'+ Support: mad@cern.ch, ', output[1]
        m.input('foo = 3;')
        assert len(output) == 1
        m.input('foo = 3;')
        assert len(output) == 2
        assert output[1] == b'++++++ info: foo redefined\n'
    assert len(output) == 3
    assert b'+          MAD-X finished normally ' in output[2]


def test_quit(mad):
    mad.quit()
    assert mad._process.returncode is not None
    assert not bool(mad)
    with raises(RuntimeError):
        mad.input(';')


@mark.xfail(
    sys.platform != 'linux',
    reason='Output is sometimes garbled on MacOS and windows.',
)
def test_context_manager():
    output = []
    with Madx(stdout=output.append) as m:
        m.input('foo = 3;')
        assert m.globals.foo == 3
    assert b'+          MAD-X finished normally ' in output[-1]
    assert not bool(m)
    with raises(RuntimeError):
        m.input(';')


def test_command_log():
    """Check that the command log contains all input commands."""
    # create a new Madx instance that uses the history feature:
    history_filename = '_test_madx.madx.tmp'
    try:
        # feed some input lines and compare with history file:
        lines = """
            l = 5;
            f = 200;

            fodo: sequence, refer=entry, l=100;
                QF: quadrupole, l=5, at= 0, k1= 1/(f*l);
                QD: quadrupole, l=5, at=50, k1=-1/(f*l);
            endsequence;

            beam, particle=proton, energy=2;
            use, sequence=fodo;
        """.splitlines()
        lines = [line.strip() for line in lines if line.strip()]
        with Madx(command_log=history_filename) as mad:
            for line in lines:
                mad.input(line)
            with open(history_filename) as history_file:
                history = history_file.read()
            assert history.strip() == '\n'.join(lines).strip()
    finally:
        # remove history file
        os.remove(history_filename)


def test_append_semicolon():
    """Check that semicolon is automatically appended to input() text."""
    # Regression test for #73
    log = []
    with Madx(command_log=log.append) as mad:
        mad.input('a = 0')
        mad.input('b = 1')
        assert log == ['a = 0;', 'b = 1;']
        assert mad.globals.a == 0
        assert mad.globals.b == 1


def test_call_and_chdir(mad):
    folder = os.path.abspath(os.path.dirname(__file__))
    parent = os.path.dirname(folder)
    getcwd = mad._libmadx.getcwd
    g = mad.globals

    mad.chdir(folder)
    assert normalize(getcwd()) == normalize(folder)
    mad.call('answer_42.madx')
    assert g.answer == 42

    with mad.chdir('..'):
        assert normalize(getcwd()) == normalize(parent)
        mad.call('test/answer_43.madx')
        assert g.answer == 43
        mad.call('test/answer_call42.madx', True)
        assert g.answer == 42

    assert normalize(getcwd()) == normalize(folder)
    mad.call('answer_43.madx')
    assert g.answer == 43

    mad.chdir('..')
    assert normalize(getcwd()) == normalize(parent)


def _check_twiss(mad, seq_name):
    beam = 'ex=1, ey=2, particle=electron, sequence={0};'.format(seq_name)
    mad.command.beam(beam)
    mad.use(seq_name)
    initial = dict(alfx=0.5, alfy=1.5,
                   betx=2.5, bety=3.5)
    twiss = mad.twiss(sequence=seq_name, **initial)
    # Check initial values:
    assert twiss['alfx'][0] == approx(initial['alfx'])
    assert twiss['alfy'][0] == approx(initial['alfy'])
    assert twiss['betx'][0] == approx(initial['betx'])
    assert twiss['bety'][0] == approx(initial['bety'])
    assert twiss.summary['ex'] == approx(1)
    assert twiss.summary['ey'] == approx(2)
    # Check that keys are all lowercase:
    for k in twiss:
        assert k == k.lower()
    for k in twiss.summary:
        assert k == k.lower()


def test_error(mad):
    mad.input("""
        seq: sequence, l=1;
        endsequence;
        beam;
        use, sequence=seq;
    """)
    # Errors in MAD-X must not crash, but return False instead:
    assert not mad.input('twiss;')
    assert mad.input('twiss, betx=1, bety=1;')


def test_twiss_1(mad):
    mad.input(SEQU)
    _check_twiss(mad, 's1')     # s1 can be computed at start
    _check_twiss(mad, 's1')     # s1 can be computed multiple times
    _check_twiss(mad, 's2')     # s2 can be computed after s1


def test_twiss_2(mad):
    mad.input(SEQU)
    _check_twiss(mad, 's2')     # s2 can be computed at start
    _check_twiss(mad, 's1')     # s1 can be computed after s2


def test_twiss_with_range(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s1;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s1')
    params = dict(alfx=0.5, alfy=1.5,
                  betx=2.5, bety=3.5,
                  sequence='s1')
    # Compute TWISS on full sequence, then on a sub-range, then again on
    # the full sequence. This checks that none of the range selections
    # have side-effects on each other:
    betx_full1 = mad.twiss(**params)['betx']
    betx_range = mad.twiss(range=('dr[2]', 'sb'), **params)['betx']
    betx_full2 = mad.twiss(**params)['betx']
    # Check that the results have the expected lengths:
    assert len(betx_full1) == 9
    assert len(betx_range) == 4
    assert len(betx_full2) == 9
    # Check numeric results. Since the first 3 elements of range and full
    # sequence are identical, equal results are expected. And non-equal
    # results afterwards.
    assert betx_range[0] == approx(betx_full1[1])      # dr:2, dr:1
    assert betx_range[1] == approx(betx_full1[2])      # qp:2, qp:1
    assert betx_range[2] == approx(betx_full1[3])      # dr:3, dr:2
    assert betx_range[3] != approx(betx_full1[4])   # sb, qp:2


def test_range_row_api(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s1;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s1')
    params = dict(alfx=0.5, alfy=1.5,
                  betx=2.5, bety=3.5,
                  sequence='s1')
    tab = mad.twiss(range=('dr[2]', 'sb'), **params)
    assert tab.range == ('dr[2]', 'sb')
    assert 'betx' in tab


def test_survey(mad):
    mad.input(SEQU)
    mad.beam()
    mad.use('s1')
    tab = mad.survey()
    assert tab._name == 'survey'
    assert 'x' in tab
    assert 'y' in tab
    assert 'z' in tab
    assert 'theta' in tab
    assert 'phi' in tab
    assert 'psi' in tab
    assert tab.x[-1] < -1
    assert tab.y == approx(0)
    assert tab.z[-1] > 7


def test_match(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s2;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s2')

    params = dict(alfx=0.5, alfy=1.5,
                  betx=2.5, bety=3.5,
                  sequence='s2')

    mad.match(constraints=[dict(range='s1$end', betx=2.0)],
              weight={'betx': 2},
              vary=['qp2->k1'],
              **params)
    twiss = mad.twiss(**params)
    val = twiss.betx[-1]
    assert val == approx(2.0, rel=1e-2)


def test_verbose(mad):
    mad.verbose(False)
    assert mad.options.echo is False
    assert mad.options.info is False
    mad.verbose(True)
    assert mad.options.echo is True
    assert mad.options.info is True


def test_active_sequence(mad):
    mad.input(SEQU)
    mad.command.beam('ex=1, ey=2, particle=electron, sequence=s1;')
    mad.use('s1')
    assert mad.sequence() == 's1'
    mad.beam()
    mad.use('s2')
    assert mad.sequence().name == 's2'


def test_get_sequence(mad):
    mad.input(SEQU)
    with raises(KeyError):
        mad.sequence['sN']
    s1 = mad.sequence['s1']
    assert s1.name == 's1'
    seqs = mad.sequence
    assert set(seqs) == {'s1', 's2'}


def test_eval(mad):
    mad.input(SEQU)
    assert mad.eval(True) is True
    assert mad.eval(13) == 13
    assert mad.eval(1.3) == 1.3
    assert mad.eval([2, True, 'QP_K1']) == [2, True, 2.0]
    assert mad.eval("1/QP_K1") == approx(0.5)


def test_eval_functions(mad):
    assert mad.eval("sin(1.0)") == approx(np.sin(1.0))
    assert mad.eval("cos(1.0)") == approx(np.cos(1.0))

    mad.input("""
        mqf.k1 =  0.3037241107;
        mqd.k1 = -0.3037241107;

        fodo: sequence, l=10, refer=entry;
        mqf: quadrupole, at=0, l=1, k1:=mqf.k1;
        dff: drift,      at=1, l=4;
        mqd: quadrupole, at=5, l=1, k1:=mqd.k1;
        dfd: drift,      at=6, l=4;
        endsequence;

        beam;
        use, sequence=fodo;
        twiss, sequence=fodo, x=0.1;
    """)

    elems = mad.sequence.fodo.expanded_elements
    twiss = mad.table.twiss

    mad.input("mqf_x = table(twiss, mqf, x);")
    assert mad.eval("table(twiss, mqf, x)") \
        == twiss.row(elems.index('mqf')).x \
        == mad.globals.mqf_x


def test_globals(mad):
    g = mad.globals
    # Membership:
    assert 'FOO' not in g
    # Setting values:
    g['FOO'] = 2
    assert 'FOO' in g
    assert g['FOO'] == 2
    assert mad.eval('FOO') == 2
    # Re-setting values:
    g['FOO'] = 3
    assert mad.eval('FOO') == 3
    # Setting expressions:
    g['BAR'] = '3*foo'
    assert mad.eval('BAR') == 9
    g['FOO'] = 4
    assert mad.eval('BAR') == 12
    assert g.defs.bar == "3*foo"
    assert g.cmdpar.bar.definition == "3*foo"
    # attribute access:
    g.bar = 42
    assert g.defs.bar == 42
    assert g.cmdpar.bar.definition == 42
    assert g.BAR == 42
    # repr
    assert "'bar': 42.0" in str(g)
    with raises(NotImplementedError):
        del g['bar']
    with raises(NotImplementedError):
        del g.bar
    assert g.bar == 42     # still there
    assert 'bar' in list(g)
    assert 'foo' in list(g)
    # assert list(g) == list(g.defs)
    # assert list(g) == list(g.cmdpar)
    assert len(g) == len(list(g))
    assert len(g.defs) == len(list(g.defs))
    assert len(g.cmdpar) == len(list(g.cmdpar))


def test_elements(mad):
    mad.input(SEQU)
    assert 'sb' in mad.elements
    assert 'sb' in list(mad.elements)
    assert 'foobar' not in mad.elements
    assert mad.elements['sb']['angle'] == approx(3.14/4)
    idx = mad.elements.index('qp1')
    elem = mad.elements[idx]
    assert elem['k1'] == 3


def test_sequence_map(mad):
    mad.input(SEQU)
    seq = mad.sequence
    assert len(seq) == 2
    assert set(seq) == {'s1', 's2'}
    assert 's1' in seq
    assert 's3' not in seq
    assert hasattr(seq, 's1')
    assert not hasattr(seq, 's3')
    assert seq.s1.name == 's1'
    assert seq.s2.name == 's2'
    with raises(AttributeError):
        seq.s3


def test_table_map(mad):
    mad.input(SEQU)
    mad.beam()
    mad.use('s2')
    mad.survey(sequence='s2')
    tab = mad.table
    assert 'survey' in list(tab)
    assert 'survey' in tab
    assert 'foobar' not in tab
    assert len(tab) == len(list(tab))
    with raises(AttributeError):
        tab.foobar


def test_sequence(mad):
    mad.input(SEQU)
    s1 = mad.sequence.s1
    assert str(s1) == '<Sequence: s1>'
    assert s1 == mad.sequence.s1
    assert s1 == 's1'
    assert s1 != mad.sequence.s2
    assert s1 != 's2'
    with raises(RuntimeError):
        s1.beam
    with raises(RuntimeError):
        s1.twiss_table
    with raises(RuntimeError):
        s1.twiss_table_name
    assert not s1.has_beam
    assert not s1.is_expanded
    s1.expand()
    assert s1.has_beam
    assert s1.is_expanded
    s1.expand()     # idempotent
    assert s1.has_beam
    assert s1.is_expanded
    initial = dict(alfx=0.5, alfy=1.5,
                   betx=2.5, bety=3.5)
    mad.twiss(sequence='s1', sectormap=True,
              table='my_twiss', **initial)
    # Now works:
    assert s1.beam.particle == 'positron'
    assert s1.twiss_table_name == 'my_twiss'
    assert s1.twiss_table.betx[0] == 2.5
    assert s1.element_names() == [
        's1$start',
        'dr', 'qp', 'dr[2]', 'qp[2]', 'dr[3]', 'sb', 'dr[4]',
        's1$end',
    ]
    assert s1.expanded_element_names() == s1.element_names()
    assert len(s1.element_names()) == len(s1.element_positions())
    assert s1.element_positions() == [
        0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 8.0]
    assert s1.expanded_element_positions() == [
        0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 8.0]

    assert s1.elements[0].name == 's1$start'
    assert s1.elements[-1].name == 's1$end'
    assert s1.elements[-1].index == len(s1.elements)-1
    assert s1.elements[3].index == 3

    assert s1.elements.index('#s') == 0
    assert s1.elements.index('#e') == len(s1.elements)-1
    assert s1.elements.index('sb') == 6
    assert s1.length == 8.0


def test_sequence_get_elements_s1(mad):
    mad.input(SEQU)
    s1 = mad.sequence.s1.elements
    qp1 = s1['qp[1]']
    qp2 = s1['qp[2]']
    sb1 = s1['sb[1]']
    assert s1.index('qp') < s1.index('qp[2]')
    assert s1.index('qp[2]') < s1.index('sb')
    assert qp1['at'] == approx(1.5)
    assert qp2['at'] == approx(3.5)
    assert sb1['at'] == approx(6)
    assert qp1.position == approx(1)
    assert qp2.position == approx(3)
    assert sb1.position == approx(5)
    assert qp1['l'] == approx(1)
    assert qp2['l'] == approx(1)
    assert sb1['l'] == approx(2)
    assert float(qp1['k1']) == approx(2)
    assert float(qp2['k1']) == approx(2)
    assert float(sb1['angle']) == approx(3.14/4)
    assert qp1.cmdpar.k1.expr.lower() == "qp_k1"


def test_sequence_get_elements_s2(mad):
    mad.input(SEQU)
    s2 = mad.sequence.s2.elements
    qp1 = s2['qp1[1]']
    qp2 = s2['qp2[1]']
    assert s2.index('qp1') < s2.index('qp2')
    assert qp1['at'] == approx(0)
    assert qp2['at'] == approx(1)
    assert qp1['l'] == approx(1)
    assert qp2['l'] == approx(2)
    assert float(qp1['k1']) == approx(3)
    assert float(qp2['k1']) == approx(2)


# def test_sequence_get_expanded_elements():


def test_crash(mad):
    """Check that a RuntimeError is raised in case MAD-X crashes."""
    assert bool(mad)
    # a.t.m. MAD-X crashes on this input, because the L (length)
    # parametere is missing:
    raises(RuntimeError, mad.input, 'XXX: sequence;')
    assert not bool(mad)


def test_sequence_elements(mad):
    mad.input(SEQU)
    elements = mad.sequence['s1'].elements
    iqp2 = elements.index('qp[2]')
    qp1 = elements['qp[1]']
    qp2 = elements[iqp2]
    assert qp1['at'] == approx(1.5)
    assert qp2['at'] == approx(3.5)
    assert qp1.position == approx(1)
    assert qp2.position == approx(3)
    assert iqp2 == elements.at(3.1)


def test_sequence_expanded_elements(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s1;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s1')
    elements = mad.sequence['s1'].expanded_elements
    iqp2 = elements.index('qp[2]')
    qp1 = elements['qp[1]']
    qp2 = elements[iqp2]
    assert qp1['at'] == approx(1.5)
    assert qp2['at'] == approx(3.5)
    assert qp1.position == approx(1)
    assert qp2.position == approx(3)
    assert iqp2 == elements.at(3.1)


def test_element_inform(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s1;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s1')
    elem = mad.sequence.s1.expanded_elements['qp']
    assert {
        name for name in elem
        if elem.cmdpar[name].inform
    } == {'k1', 'l', 'at'}


def test_table(mad):
    beam = 'ex=1, ey=2, particle=electron, sequence=s1;'
    mad.input(SEQU)
    mad.command.beam(beam)
    mad.use('s1')
    initial = dict(alfx=0.5, alfy=1.5,
                   betx=2.5, bety=3.5)
    twiss = mad.twiss(sequence='s1', sectormap=True, **initial)
    sector = mad.table.sectortable

    assert str(twiss).startswith("<Table 'twiss': ")
    assert str(sector).startswith("<Table 'sectortable': ")

    assert 'betx' in twiss
    assert 't111' in sector
    assert 't111' not in twiss
    assert 'betx' not in sector

    assert len(twiss) == len(list(twiss))
    assert set(twiss) == set(twiss[0])
    assert twiss.s[5] == twiss[5].s
    assert twiss.s[-1] == twiss[-1].s

    copy = twiss.copy()
    assert copy['betx'] == approx(twiss.betx)
    assert set(copy) == set(twiss)
    copy = twiss.copy(['betx'])
    assert set(copy) == {'betx'}

    ALL = slice(None)

    assert sector.tmat(0).shape == (6, 6, 6)
    assert_allclose(sector.tmat(ALL)[0, 0, 0, :], sector.t111)
    assert_allclose(sector.tmat(ALL)[1, 5, 3, :], sector.t264)
    assert_allclose(sector.tmat(ALL)[3, 0, 3, :], sector.t414)
    assert_allclose(sector.tmat(ALL)[4, 4, 4, :], sector.t555)

    assert_allclose(sector.rmat(ALL)[0, 0, :], sector.r11)
    assert_allclose(sector.rmat(ALL)[1, 5, :], sector.r26)
    assert_allclose(sector.rmat(ALL)[3, 0, :], sector.r41)
    assert_allclose(sector.rmat(ALL)[4, 4, :], sector.r55)

    assert_allclose(sector.kvec(ALL)[0, :], sector.k1)
    assert_allclose(sector.kvec(ALL)[1, :], sector.k2)
    assert_allclose(sector.kvec(ALL)[3, :], sector.k4)
    assert_allclose(sector.kvec(ALL)[4, :], sector.k5)

    r = mad.sectortable()[:, :6, :6]
    k = mad.sectortable()[:, 6, :6]
    t = mad.sectortable2()

    num_elems = len(mad.sequence.s1.elements)
    assert t.shape == (num_elems, 6, 6, 6)
    assert r.shape == (num_elems, 6, 6)
    assert k.shape == (num_elems, 6)

    assert_allclose(t[:, 0, 0, 0], sector.t111)
    assert_allclose(t[:, 1, 5, 3], sector.t264)
    assert_allclose(t[:, 3, 0, 3], sector.t414)
    assert_allclose(t[:, 4, 4, 4], sector.t555)

    assert_allclose(r[:, 0, 0], sector.r11)
    assert_allclose(r[:, 1, 5], sector.r26)
    assert_allclose(r[:, 3, 0], sector.r41)
    assert_allclose(r[:, 4, 4], sector.r55)

    assert_allclose(k[:, 0], sector.k1)
    assert_allclose(k[:, 1], sector.k2)
    assert_allclose(k[:, 3], sector.k4)
    assert_allclose(k[:, 4], sector.k5)


def test_selected_columns(mad, lib):
    mad.input(SEQU)
    mad.command.beam()
    mad.use('s1')

    mad.select(flag='twiss', column=['s', 'x', 'y'])
    table = mad.twiss(sequence='s1', betx=1, bety=1)
    assert set(table) > {'s', 'x', 'y', 'betx', 'bety'}
    assert set(table.copy()) > {'s', 'x', 'y', 'betx', 'bety'}
    assert table.selected_columns() == ['s', 'x', 'y']
    assert table.selection().col_names() == ['s', 'x', 'y']
    assert table.selection().copy().keys() == {'s', 'x', 'y'}

    mad.select(flag='twiss', clear=True)
    mad.select(flag='twiss', column=['betx', 'bety'])
    lib.apply_table_selections('twiss')
    table = mad.table.twiss
    assert set(table) > {'s', 'x', 'y', 'betx', 'bety'}
    assert set(table.copy()) > {'s', 'x', 'y', 'betx', 'bety'}
    assert table.selected_columns() == ['betx', 'bety']
    assert table.selection().col_names() == ['betx', 'bety']
    assert table.selection().copy().keys() == {'betx', 'bety'}


def test_table_selected_rows(mad, lib):
    mad.input(SEQU)
    mad.command.beam()
    mad.use('s1')

    def check_selection(table, name):
        assert_equal(
            table.column(name, rows='selected'),
            table[name][table.selected_rows()])
        assert_equal(
            table.column(name, rows='selected'),
            table.selection()[name])

    mad.select(flag='twiss', class_='quadrupole')
    table = mad.twiss(sequence='s1', betx=1, bety=1)
    assert table.selected_rows() == [2, 4]
    check_selection(table, 'alfx')
    check_selection(table, 'alfy')
    check_selection(table, 'betx')
    check_selection(table, 'bety')

    mad.select(flag='twiss', clear=True)
    mad.select(flag='twiss', class_='drift')
    lib.apply_table_selections('twiss')
    table = mad.table.twiss
    assert table.selected_rows() == [1, 3, 5, 7]
    check_selection(table, 'alfx')
    check_selection(table, 'alfy')
    check_selection(table, 'betx')
    check_selection(table, 'bety')


def test_table_selected_rows_mask(mad, lib):
    mad.input(SEQU)
    mad.command.beam()
    mad.use('s1')

    mad.select(flag='twiss', class_='quadrupole')
    table = mad.twiss(sequence='s1', betx=1, bety=1)
    mask = lib.get_table_selected_rows_mask('twiss')
    assert mask.shape == (len(mad.sequence.s1.expanded_elements), )
    assert_equal(mask.nonzero(), (table.selected_rows(), ))


def test_attr(mad):
    assert hasattr(mad, 'constraint')
    assert hasattr(mad, 'constraint_')
    assert hasattr(mad, 'global_')
    assert not hasattr(mad, 'foobar')
    assert not hasattr(mad, '_constraint')


def test_expr(mad):
    g = mad.globals
    vars = mad.expr_vars
    g.foo = 1
    g.bar = 2
    assert set(vars('foo')) == {'foo'}
    assert set(vars('(foo) * sin(2*pi*bar)')) == {'foo', 'bar'}


def test_command(mad):
    mad.input(SEQU)
    twiss = mad.command.twiss
    sbend = mad.elements.sb
    clone = sbend.clone('foobar', angle="pi/5", l=1)

    assert 'betx=0' in str(twiss)
    assert 'angle=' in str(sbend)
    assert 'tilt' in sbend
    assert sbend.tilt == 0
    assert len(sbend) == len(list(sbend))
    assert 'tilt' in list(sbend)

    assert clone.name == 'foobar'
    assert clone.base_type.name == 'sbend'
    assert clone.parent.name == 'sb'
    assert clone.defs.angle == 'pi / 5'
    assert clone.angle == approx(0.6283185307179586)
    assert len(clone) == len(sbend)

    assert 'angle=0.628' in str(clone)
    assert 'tilt' not in str(clone)
    clone.angle = 0.125
    clone = mad.elements.foobar            # need to update cache
    assert clone.angle == 0.125
    assert len(twiss) == len(list(twiss))
    assert 'betx' in list(twiss)

    assert clone.angle != approx(clone.parent.angle)
    del clone.angle
    clone = mad.elements.foobar            # need to update cache
    assert clone.angle == clone.parent.angle

    with raises(AttributeError):
        clone.missing_attribute

    with raises(NotImplementedError):
        del twiss['betx']
    with raises(NotImplementedError):
        del clone.base_type.angle


def test_array_attribute(mad):
    mad.globals.nine = 9
    clone = mad.elements.multipole.clone('foo', knl=[0, 'nine/3', 4])
    knl = clone.knl
    assert knl[0] == 0
    assert knl[1] == 3
    assert knl[2] == 4
    assert len(knl) == 3
    assert list(knl) == [0.0, 3.0, 4.0]
    assert str(knl) == '[0.0, 3.0, 4.0]'
    knl[1] = '3*nine'
    assert mad.elements.foo.defs.knl[1] == '3 * nine'
    assert mad.elements.foo.knl[1] == 27


def test_array_attribute_comparison(mad):
    mad.globals.nine = 9
    foo = mad.elements.multipole.clone('foo', knl=[0, 5, 10])

    bar_eq = mad.elements.multipole.clone('bar_eq', knl=[0, 5, 10])
    bar_gt = mad.elements.multipole.clone('bar_gt', knl=[0, 6, 10])
    bar_lt = mad.elements.multipole.clone('bar_lt', knl=[0, 5, 'nine'])

    knl = foo.knl
    knl_eq = bar_eq.knl
    knl_gt = bar_gt.knl
    knl_lt = bar_lt.knl

    assert knl == knl_eq
    assert not (knl == knl_gt)
    assert not (knl == knl_lt)

    assert not (knl < knl_eq)
    assert knl < knl_gt
    assert not (knl < knl_lt)

    assert knl <= knl_eq
    assert knl <= knl_gt
    assert not (knl <= knl_lt)

    assert not (knl > knl_eq)
    assert not (knl > knl_gt)
    assert knl > knl_lt

    assert knl >= knl_eq
    assert not (knl >= knl_gt)
    assert knl >= knl_lt


def test_command_map(mad):
    command = mad.command
    assert 'match' in command
    assert 'sbend' in command

    assert 'foooo' not in command

    assert 'match' in list(command)
    assert len(command) == len(list(command))
    assert 'match' in str(command)
    assert 'sbend' in str(command)

    assert 'sbend' in mad.base_types
    assert 'match' not in mad.base_types


def test_comments(mad):
    var = mad.globals
    mad('x = 1; ! x = 2;')
    assert var.x == 1
    mad('x = 2; // x = 3;')
    assert var.x == 2
    mad('x = 3; /* x = 4; */')
    assert var.x == 3
    mad('/* x = 3; */ x = 4;')
    assert var.x == 4
    mad('x = 5; ! /* */ x = 6;')
    assert var.x == 5
    mad('x = 5; /* ! */ x = 6;')
    assert var.x == 6


def test_multiline_input(mad):
    var = mad.globals
    mad('''
        x = 1;
        y = 2;
    ''')
    assert var.x, 1
    assert var.y, 2
    mad('''
        x = /* 3;
        y =*/ 4;
    ''')
    assert var.x == 4
    assert var.y == 2
    mad('''
        x = 1; /*  */ x = 2;
        */ if (x == 1) {
            x = 3;
        }
    ''')
    assert var.x == 2
    mad('''
        x = 1; /* x = 2;
        */ if (x == 1) {
            x = 3;
        }
    ''')
    assert var.x == 3


def test_errors(mad):
    mad.input(SEQU)
    mad.beam()
    mad.use(sequence='s1')
    mad.select(flag='error', range='qp')
    dkn = [1e-6, 2e-6, 3e-6]
    dks = [4e-6, 5e-6, 6e-6]
    mad.efcomp(dkn=dkn, dks=dks)
    mad.ealign(dx=1e-3, dy=-4e-3)
    fd = mad.sequence['s1'].expanded_elements['qp'].field_errors
    al = mad.sequence['s1'].expanded_elements['qp'].align_errors
    expected_dkn = np.hstack((dkn, np.zeros(len(fd.dkn) - len(dkn))))
    expected_dks = np.hstack((dks, np.zeros(len(fd.dks) - len(dks))))
    assert_allclose(fd.dkn, expected_dkn)
    assert_allclose(fd.dks, expected_dks)
    assert_allclose(al.dx, 1e-3)
    assert_allclose(al.dy, -4e-3)


def test_subsequence(mad):
    mad.input("""
    d1: RBEND, l=0.1, angle=0.1;

    seq1: sequence, l=0.1;
        d1.1: d1, at=0.05;
    endsequence;

    seq2: sequence, l=0.2;
        seq1, at=0.05;
        seq1, at=0.15;
    endsequence;
    """)
    seq2 = mad.sequence.seq2
    assert isinstance(seq2.elements['seq1'], Sequence)
    assert seq2.elements['seq1'].name == 'seq1'
    assert seq2.elements['seq1'].element_names() == \
        mad.sequence.seq1.element_names()


def test_dframe_after_use(mad):
    mad.input("""
        mqf.k1 =  0.3037241107;
        mqd.k1 = -0.3037241107;

        fodo: sequence, l=10, refer=entry;
        mqf: quadrupole, at=0, l=1, k1:=mqf.k1;
        dff: drift,      at=1, l=4;
        mqd: quadrupole, at=5, l=1, k1:=mqd.k1;
        dfd: drift,      at=6, l=4;
        endsequence;

        beam;
        use, sequence=fodo;
        twiss, sequence=fodo, x=0.1;
    """)
    index = ['#s', 'mqf', 'dff', 'mqd', 'dfd', '#e']
    names = ['fodo$start', 'mqf', 'dff', 'mqd', 'dfd', 'fodo$end']

    twiss = mad.table.twiss
    assert index == twiss.row_names()
    assert index == twiss.dframe().index.tolist()
    assert names == twiss.dframe(index='name').index.tolist()

    mad.use(sequence='fodo')

    twiss = mad.table.twiss

    # Should still work:
    assert names == twiss.dframe(index='name').index.tolist()

    # The following assert demonstrates the current behaviour and is
    # meant to detect if the MAD-X implementation changes. It may lead
    # to crashes or change in the future. In that case, please remove
    # this line. It does not represent desired behaviour!
    assert mad.table.twiss.row_names() == \
        ['#s', '#e', 'dfd', 'mqd', 'dff', 'mqf']



"""
Simple RPC module for libmadx.

This module is used to execute several instances of the `libmadx` module
in remote processes and communicate with them via remote procedure calls
(RPC). Use the :meth:`LibMadxClient.spawn` to create a new instance.

The remote backend is needed due to the fact that cpymad.libmadx is a low
level binding to the MAD-X library which in turn uses global variables.
This means that the cpymad.libmadx module has to be loaded within remote
processes in order to deal with several isolated instances of MAD-X in
parallel. Furthermore, running MAD-X in a separate process shields the main
python process from crashes due to incorrect use of MAD-X.
"""

from __future__ import absolute_import

from minrpc.client import Client, RemoteProcessCrashed, RemoteProcessClosed


__all__ = [
    'LibMadxClient',
    'RemoteProcessCrashed',
    'RemoteProcessClosed',
]


class LibMadxClient(Client):

    """
    Specialized client for boxing :mod:`cpymad.libmadx` function calls.
    """

    def close(self):
        """Finalize libmadx if it is running."""
        try:
            if self.libmadx.is_started():
                self.libmadx.finish()
        except (RemoteProcessClosed, RemoteProcessCrashed):
            pass
        super(LibMadxClient, self).close()

    @property
    def libmadx(self):
        return self.get_module('cpymad.libmadx')



"""
Implementation of a simple LL(1) parser.
"""
from copy import deepcopy


def fix_point(func, x, **kwargs):
    """
    Repeatedly apply ``x := f(x, **kwargs)`` until encountering a fix
    point, i.e. ``x == f(x)`` and return the resulting fix point ``x``.
    """
    while True:
        x_new = func(x, **kwargs)
        if x_new == x:
            return x
        else:
            x = x_new


def extend_empty_sets(empty: dict, grammar: dict) -> dict:
    """
    Determine which nonterminals of an LL(1) grammar can be empty.
    Must be applied repeatedly until converging.

    :param empty: nonterminal -> bool
    :param grammar: nonterminal -> [productions...]
    :returns: Extended copy of ``empty``
    """
    return {
        symbol: empty.get(symbol) or any(
            all(empty.get(p) for p in production)
            for production in productions
        )
        for symbol, productions in grammar.items()
    }


def extend_firsts_sets(firsts, *, empty, grammar):
    """
    Determine the FIRST sets of an LL(1) grammar, i.e. the set of terminals
    that a given symbol can start with.

    Must be applied repeatedly until converging.

    :param firsts: terminal|nonterminal -> set[terminal]
    :param empty: nonterminal -> bool
    :param grammar: nonterminal -> [productions...]
    :returns: Extended copy of ``firsts``
    """
    firsts = deepcopy(firsts)
    for symbol, productions in grammar.items():
        for production in productions:
            for i, n in enumerate(production):
                extend_parse_table(symbol, firsts[symbol], {
                    t: production[i + 1:][::-1] + p
                    for t, p in firsts[n].items()
                })
                if not empty[n]:
                    break
    return firsts


def extend_follow_sets(follow, *, firsts, empty, grammar):
    """
    Determine the FOLLOW sets of an LL(1) grammar, i.e. the set of terminals
    that a given symbol can be followed by if it may be empty.

    Must be applied repeatedly until converging.

    :param follow: terminal|nonterminal -> set[terminal]
    :param firsts: terminal|nonterminal -> set[terminal]
    :param empty: nonterminal -> bool
    :param grammar: nonterminal -> [productions...]
    :returns: Extended copy of ``follow``
    """
    follow = deepcopy(follow)
    for symbol, productions in grammar.items():
        for production in productions:
            for i, p1 in enumerate(production):
                for p2 in production[i + 1:]:
                    follow[p1] |= set(firsts[p2])
                    if not empty[p2]:
                        break
                else:
                    follow[p1] |= follow[symbol]
    return follow


def create_parse_table(terminals, grammar, start):
    """
    Create an LL(1) parsing table.

    :param terminals: list of terminal symbols
    :param grammar: nonterminal -> [productions...]
    :param start: nonterminal start symbol
    :returns: parse table nonterminal -> {terminal -> production}
    """
    empty = fix_point(extend_empty_sets, {}, grammar=grammar)
    empty.update({t: False for t in terminals})
    table = {**{t: {t: []} for t in terminals}, **{n: {} for n in grammar}}
    table = fix_point(
        extend_firsts_sets,
        table,
        empty=empty,
        grammar=grammar)
    follow = {
        **{t: set() for t in terminals},
        **{n: set() for n in grammar},
    }
    follow = fix_point(
        extend_follow_sets, follow,
        firsts=table, empty=empty, grammar=grammar)

    for symbol, terminals in follow.items():
        if empty[symbol]:
            extend_parse_table(symbol, table[symbol], {
                t: None
                for t in terminals
            })

    return table


def extend_parse_table(symbol, old, new):
    """
    Merge parse tables ``old`` and ``new`` without silently overriding
    duplicate entries. Raises a ``ValueError`` if ``new`` contains entries
    that were defined differently in ``old``.
    """
    for key in old.keys() & new.keys():
        if old[key] != new[key]:
            raise ValueError((
                "Grammar is ambiguous. Duplicate entry "
                "in parse table: <{}, {}> -> {} or {}"
            ).format(symbol, key, old[key], new[key]))
    old.update(new)


class Parser:

    """
    LL(1) syntax checker.

    :param terminals: list of terminal symbols
    :param grammar: nonterminal -> [productions...]
    :param start: nonterminal start symbol
    """

    def __init__(self, terminals, grammar, start):
        self.table = table = create_parse_table(
            terminals, grammar, start)
        # precompule rules lookup:
        self.rules = {symbol: {} for symbol in table}
        for symbol, rules in table.items():
            self.rules[symbol].update({
                t: p and [self.rules[n] for n in p]
                for t, p in rules.items()
            })
        self.start = self.rules[start]

    def parse(self, tokens):
        """
        Verifies the grammar.

        :param list tokens: list of tokens (terminals) in input order
        :returns: nothing if successful
        :raises: ValueError
        """
        tokens = list(reversed(tokens))
        stack = [self.start]
        try:
            while stack:
                token = tokens[-1]
                rules = stack.pop()
                more = rules[token.type]
                if more is not None:
                    stack.extend(more)
                    tokens.pop()
        except KeyError:
            raise ValueError(
                ("Unexpected {} in:\n"
                 "    {!r}\n"
                 "     ").format(token.type, token.expr)
                + ' ' * token.start
                + '^' * max(token.length, 1)
            ) from None



# cython: embedsignature=True
# cython: annotation_typing=False
# cython: language_level=3
"""
Low level cython binding to MAD-X.

CAUTION: End users should usually not import this module directly! Use
:class:`cpymad.madx.Madx` instead.

- The API of this module is considered private, i.e. it may change between
  versions without further notice.
- Importing this module means loading MAD-X directly into the process space.
  This means that any crash in the (sometimes fragile) MAD-X interpreter will
  crash the importing process with it.
- All functions in this module operate on a shared global state.
- this module exposes a very C-ish API that is not convenient to work with.
"""

import os
import ctypes
import numpy as np      # Import the Python-level symbols of numpy

# Import a large-enough integer type to hold pointer, see also:
# http://grokbase.com/t/gg/cython-users/134b21rga8/passing-callback-pointers-to-python-and-back
cdef extern from "pyport.h":
    ctypedef int Py_intptr_t

cdef extern from "string.h" nogil:
    char* strstr(char* s1, char* s2)
    char* strchr(char* s, int c)

from cpymad.types import Constraint, Parameter, AlignError, FieldError, PhaseError
from cpymad.util import name_to_internal, name_from_internal, normalize_range_name
cimport cpymad.clibmadx as clib


# Remember whether start() was called
_madx_started = False


# Python-level binding to libmadx:
__all__ = [
    # MAD-X version
    'get_version_number',
    'get_version_date',

    # control the interpreter
    'is_started',
    'start',
    'finish',
    'input',
    'eval',

    # Globals
    'get_var',
    'num_globals',
    'get_globals',
    'get_var_type',

    'get_options',

    # iterate sequences
    'sequence_exists',
    'get_sequence_names',
    'get_sequence_count',
    'get_active_sequence_name',

    # sequence access
    'get_sequence_twiss_table_name',
    'get_sequence_beam',
    'is_sequence_expanded',

    # iterate tables
    'table_exists',
    'get_table_names',
    'get_table_count',

    # table access
    'get_table_summary',
    'get_table_column_names',
    'get_table_column_count',
    'get_table_column',
    'get_table_row',
    'get_table_row_count',
    'get_table_row_names',

    'get_table_selected_rows',
    'get_table_selected_rows_mask',
    'apply_table_selections',

    # sequence element list access
    'get_element',
    'get_element_name',
    'get_element_positions',
    'get_element_names',
    'get_element_index',
    'get_element_index_by_position',
    'get_element_count',

    # expanded sequence element access
    'get_expanded_element',
    'get_expanded_element_name',
    'get_expanded_element_positions',
    'get_expanded_element_names',
    'get_expanded_element_index',
    'get_expanded_element_index_by_position',
    'get_expanded_element_count',

    # global elements
    'get_global_element',
    'get_global_element_name',
    'get_global_element_index',
    'get_global_element_count',

    # element base types
    'get_base_type_names',

    # defined commands
    'get_defined_command',
    'get_defined_command_names',

    # imported from 'os' for convenience in madx.Madx and should not be
    # considered part of the public interface:
    'getcwd',
]


def getcwd() -> str:
    """Return the current working directory."""
    return os.getcwd()


def _get_rightmost_word(sentence: str) -> str:
    """Get the work right of the rightmost space character."""
    return sentence.rsplit(' ', 1)[-1]


def get_version_number() -> str:
    """
    Get the version number of loaded MAD-X interpreter.

    :returns: full version number
    """
    return _get_rightmost_word(_str(clib.version_name))


def get_version_date() -> str:
    """
    Get the release date of loaded MAD-X interpreter.

    :returns: release date in YYYY.MM.DD format
    """
    return _get_rightmost_word(_str(clib.version_date))


def is_started() -> bool:
    """
    Check whether MAD-X has been initialized.

    :returns: whether :func:`start` was called without matching :func:`finish`
    """
    return _madx_started


def start() -> None:
    """
    Initialize MAD-X.
    """
    clib.madx_start()
    global _madx_started
    _madx_started = True


def finish() -> None:
    """
    Cleanup MAD-X.
    """
    clib.madx_finish()
    global _madx_started
    _madx_started = False


def input(cmd: str) -> bool:
    """
    Pass one input command to MAD-X.

    :param str cmd: command to be executed by the MAD-X interpreter
    :returns: success status, whether the command has completed without error
    """
    cmd = cmd.rstrip().rstrip(';') + ';'
    cdef bytes _cmd = _cstr(cmd)
    cdef char* _pch = _cmd
    cdef int error = 0
    with nogil:
        _strip_comments(_pch)
        clib.supp_lt(_pch, 0)
        clib.stolower_nq(_pch)
        clib.pro_input(_pch)
        error = clib.geterrorflag()
        clib.clearerrorflag()
    return not error


def get_var(name: str) -> Parameter:
    """
    Get the value of a global variable.
    """
    cdef clib.variable* var = _get_var(name)
    cdef int index = clib.name_list_pos(var.name, clib.variable_list.list)
    cdef int inform = index >= clib.start_var
    if var.type == clib.VAR_TYPE_STRING:
        return Parameter(
            name, _str(var.string), None,
            dtype=clib.PARAM_TYPE_STRING, inform=inform, var_type=var.type)
    cdef double value = clib.variable_value(var)
    cdef int typeid
    if var.val_type == clib.VAL_TYPE_INTEGER:
        typeid = clib.PARAM_TYPE_INTEGER
    else:
        typeid = clib.PARAM_TYPE_DOUBLE
    return Parameter(
        name, *_expr(var.expr, value, typeid),
        dtype=typeid, inform=inform, var_type=var.type)


def get_var_type(name: str) -> int:
    """
    Get the type of the variable:

        0   constant
        1   direct
        2   deferred
        3   string
    """
    return _get_var(name).type


def get_options() -> dict:
    """Get the current option values."""
    return _parse_command(clib.options)


def num_globals() -> int:
    """
    Return the number of global variables.
    """
    return clib.variable_list.curr


def get_globals() -> list:
    """
    Get a list of names of all global variables.
    """
    return _name_list(clib.variable_list.list)


def sequence_exists(sequence_name: str) -> bool:
    """
    Check if the sequence exists.

    :param str sequence_name: sequence name
    :returns: True if the sequence exists
    """
    try:
        _find_sequence(sequence_name)
        return True
    except ValueError:
        return False


def get_sequence_twiss_table_name(sequence_name: str) -> bool:
    """
    Get the last calculated twiss table for the given sequence.

    :param str sequence_name: sequence name
    :returns: twiss table name
    :raises ValueError: if the sequence name is invalid
    :raises RuntimeError: if the twiss table is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if not seq.tw_valid:
        raise RuntimeError("TWISS table invalid for {!r}".format(sequence_name))
    return _str(seq.tw_table.name)


def get_sequence_beam(sequence_name: str) -> dict:
    """
    Get the beam associated to the sequence.

    :param str sequence_name: sequence name
    :returns: beam properties as set with the BEAM command (and some more)
    :raises ValueError: if the sequence name is invalid
    :raises RuntimeError: if the sequence has no associated beam
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if seq.beam is NULL or not seq.beam.beam_def:
        raise RuntimeError("No beam attached to {!r}".format(sequence_name))
    return _parse_command(seq.beam)


def get_sequence_length(sequence_name: str) -> Parameter:
    """
    Get the length associated to the sequence.

    :param str sequence_name: sequence name
    :returns: sequence length
    :raises ValueError: if the sequence name is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    length = Parameter(
            'length',
            *_expr(seq.l_expr, seq.length),
            dtype=clib.PARAM_TYPE_DOUBLE,
            inform=1)
    return length


def get_active_sequence_name() -> str:
    """
    Get the name of the active sequence.

    :returns: name of active sequence
    :raises RuntimeError: if no sequence is activated
    """
    if clib.current_sequ is NULL:
        raise RuntimeError("No active sequence!")
    return _str(clib.current_sequ.name)


def get_sequence_names() -> list:
    """
    Get a list of all sequences currently in memory.

    :returns: sequence names
    """
    cdef clib.sequence_list* seqs = clib.madextern_get_sequence_list()
    cdef int i
    return [_str(seqs.sequs[i].name)
            for i in range(seqs.curr)]


def get_sequence_count() -> int:
    """
    Get the number of all sequences currently in memory.

    :returns: number of sequences
    """
    return clib.madextern_get_sequence_list().curr


def table_exists(table_name: str) -> bool:
    """
    Check if the table exists.

    :param str table_name: table name
    :returns: True if the table exists
    """
    cdef bytes _table_name = _cstr(table_name)
    return bool(clib.table_exists(_table_name))


def get_table_names() -> list:
    """
    Return list of all table names.

    :returns: table names
    """
    return [_str(clib.table_register.names.names[i])
            for i in range(clib.table_register.names.curr)]


def get_table_count() -> int:
    """
    Return number of existing tables.

    :returns: number of tables in memory
    """
    return clib.table_register.names.curr


def get_table_summary(table_name: str) -> dict:
    """
    Get table summary.

    :param str table_name: table name
    :returns: mapping of {column: value}
    """
    cdef bytes _table_name = _cstr(table_name)
    cdef clib.char_p_array* header = clib.table_get_header(_table_name)
    cdef int i
    if header is NULL:
        raise ValueError("No summary for table: {!r}".format(table_name))
    return dict([_split_header_line(header.p[i])
                 for i in range(header.curr)])


def get_table_column_names(table_name: str, selected: bool = False) -> list:
    """
    Get a list of all column names in the table.

    :param str table_name: table name
    :param bool selected: consider only selected columns
    :returns: column names
    :raises ValueError: if the table name is invalid
    """
    cdef clib.table* table = _find_table(table_name)
    cdef clib.name_list* columns = table.columns;
    cdef int i
    # NOTE: we can't enforce lower-case on the column names here, since this
    # breaks their usage with get_table_column():
    if selected:
        indices = [table.col_out.i[i] for i in range(table.col_out.curr)]
        return [_str(table.columns.names[i]) for i in indices]
    else:
        VALID = (clib.PARAM_TYPE_INTEGER,
                 clib.PARAM_TYPE_DOUBLE,
                 clib.PARAM_TYPE_STRING)
        return [_str(columns.names[i]) for i in range(columns.curr)
                if columns.inform[i] in VALID]


def get_table_column_count(table_name: str, selected: bool = False) -> int:
    """
    Get a number of columns in the table.

    :param str table_name: table name
    :param bool selected: consider only selected columns
    :returns: number of columns
    :raises ValueError: if the table name is invalid
    """
    cdef clib.table* table = _find_table(table_name)
    if selected:
        return table.col_out.curr
    else:
        return table.columns.curr


def get_table_column(table_name: str, column_name: str, rows='all') -> np.ndarray:
    """
    Get data from the specified table.

    :param str table_name: table name
    :param str column_name: column name
    :returns: the data in the requested column
    :raises ValueError: if the column cannot be found in the table
    :raises RuntimeError: if the column has unknown type

    CAUTION: Numeric data is wrapped in numpy arrays but not copied. Make
    sure to copy all data before invoking any further MAD-X commands! This
    is done automatically for you if using libmadx in a remote service
    (pickle serialization effectively copies the data).
    """
    cdef char** char_tmp
    cdef bytes _tab_name = _cstr(table_name)
    cdef bytes _col_name = _cstr(column_name)
    cdef clib.column_info info = clib.table_get_column(_tab_name, _col_name)
    dtype = <bytes> info.datatype
    # row indices:
    if isinstance(rows, str):
        if rows == 'all':
            indices = np.arange(info.length)
        elif rows == 'selected':
            indices = get_table_selected_rows(table_name)
        else:
            raise ValueError("Invalid value for rows:", rows)
    else:
        indices = np.arange(info.length)[rows]
    # double:
    if dtype == b'i' or dtype == b'd':
        # YES, integers are internally stored as doubles in MAD-X:
        if info.length == 0:
            return np.empty(0)
        else:
            return np.ctypeslib.as_array(
                <double [:info.length]> info.data)[indices]
    # string:
    elif dtype == b'S':
        char_tmp = <char**> info.data
        return np.array([_str(char_tmp[i]) for i in indices], dtype=str)
    # invalid:
    elif dtype == b'V':
        raise ValueError("Column {!r} is not in table {!r}."
                         .format(column_name, table_name))
    # unknown:
    else:
        raise RuntimeError("Unknown datatype {!r} in column {!r}."
                           .format(_str(dtype), column_name))


def get_table_row(table_name: str, row_index: int, columns='all') -> dict:
    """
    Return row as tuple of values.
    """
    cdef clib.table* table = _find_table(table_name)

    if row_index < -table.curr or row_index >= table.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(row_index, table.curr))
    if row_index < 0:
        row_index += table.curr

    def fetch_value(col_index):
        inform = table.columns.inform[col_index]
        if inform == clib.PARAM_TYPE_INTEGER:
            return int(table.d_cols[col_index][row_index])
        elif inform == clib.PARAM_TYPE_DOUBLE:
            return table.d_cols[col_index][row_index]
        elif inform == clib.PARAM_TYPE_STRING:
            return _str(table.s_cols[col_index][row_index])
        raise RuntimeError("Unknown datatype {!r} in column {!r}."
                           .format(inform, table.columns.names[col_index]))

    if isinstance(columns, str):
        if columns == 'all':
            indices = range(table.columns.curr)
        elif columns == 'selected':
            indices = [table.col_out.i[i] for i in range(table.col_out.curr)]
            # fallback if selection missing:
            indices = indices or range(table.columns.curr)
        else:
            raise ValueError("Invalid value for columns:", columns)
    else:
        indices = []
        for col in columns:
            if isinstance(col, int):
                if col < 0 or col > table.columns.curr:
                    raise IndexError(
                        "Column index out of range: {0} (column count is {1})"
                        .format(col, table.columns.curr))
                indices.append(col)
            else:
                idx = clib.name_list_pos(_cstr(col), table.columns)
                if idx < 0:
                    raise ValueError("Column not found: {0!r}".format(col))
                indices.append(idx)

    return {
        _str(table.columns.names[i]): fetch_value(i)
        for i in indices
    }


def get_table_row_count(table_name: str) -> int:
    """
    Return total number of rows in the table.
    """
    return _find_table(table_name).curr


def get_table_row_names(table_name: str, indices='all') -> list:
    """
    Return row names for every index (row number) in the list.
    """
    cdef clib.table* table = _find_table(table_name)
    if isinstance(indices, str):
        if indices == 'all':
            indices = range(table.curr)
        elif indices == 'selected':
            indices = [table.row_out.i[i] for i in range(table.curr)]
        else:
            raise ValueError("Invalid value for indices:", indices)
    elif isinstance(indices, int):
        return _get_table_row_name(table, indices)
    return [_get_table_row_name(table, i) for i in indices]


def get_table_selected_rows(table_name: str) -> list:
    """Return list of selected row indices in table (may be empty)."""
    cdef clib.table* table = _find_table(table_name)
    return [i for i in range(table.curr) if table.row_out.i[i]]


def get_table_selected_rows_mask(table_name: str) -> np.ndarray:
    """Return boolean mask of which rows are selected in a table."""
    cdef clib.table* table = _find_table(table_name)
    return np.array(
        [table.row_out.i[i] for i in range(table.curr)],
        dtype=bool)


def apply_table_selections(table_name: str):
    """
    Apply the SELECT/DESELECT commands for table columns/rows.

    Needed as replacement for the missing ``out_table`` call for initializing
    ``t.row_out``, ``t.col_out`` if the twiss command was performed without a
    filename.
    """
    cdef clib.table* t = _find_table(table_name)
    version_info = tuple(map(int, get_version_number().split('.')))
    if version_info > (5,3,7):       # will crash before
        clib.out_table(_cstr(table_name), t, NULL)


def get_element(sequence_name: str, element_index: int) -> dict:
    """
    Return requested element in the original sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the element with the specified index
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index >= seq.nodes.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, seq.nodes.curr))
    return _get_node(seq.nodes.nodes[element_index], seq.ref_flag, seq.n_nodes > 0, seq.line)


def get_element_positions(sequence_name: str) -> list:
    """
    Get list with positions of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :returns: positions of all elements in the sequence
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef int i
    cdef clib.node** nodes = seq.nodes.nodes
    return [_get_node_entry_pos(nodes[i], seq.ref_flag, seq.n_nodes > 0)
            for i in range(seq.nodes.curr)]

def get_element_names(sequence_name: str) -> list:
    """
    Get list with the names of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :returns: names of all elements in the sequence
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef int i
    return [_node_name(seq.nodes.nodes[i])
            for i in range(seq.nodes.curr)]


def get_element_name(sequence_name: str, element_index: int) -> str:
    """
    Get list with the names of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the name of the element with the specified index
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index >= seq.nodes.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, seq.nodes.curr))
    return _node_name(seq.nodes.nodes[element_index])


def get_element_index(sequence_name: str, element_name: str) -> int:
    """
    Return index of element with specified name in the original sequence.

    :param str sequence_name: sequence name
    :param str element_name: element name
    :returns: the index of the specified element
    :raises ValueError: if the sequence or element name is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef bytes _element_name = _cstr(name_to_internal(element_name))
    cdef int index = clib.name_list_pos(_element_name, seq.nodes.list)
    if index == -1:
        raise ValueError("Element name not found: {0!r}".format(element_name))
    return index


def get_element_index_by_position(sequence_name: str, position: float) -> int:
    """
    Return index of element at specified position in the original sequence.

    :param str sequence_name: sequence name
    :param float position: position (S coordinate)
    :returns: the index of an element at that position, -1 if not found
    :raises ValueError: if the sequence or element name is invalid
    """
    # This is implemented here just for performance, but still uses a linear
    # algorithm. If you want more, you need to copy the elements over and use
    # a suitable ordered data type.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef double _position = position
    cdef clib.node* elem
    cdef double at
    for i in range(seq.nodes.curr):
        elem = seq.nodes.nodes[i]
        at = _get_node_entry_pos(elem, seq.ref_flag, seq.n_nodes > 0)
        if _position >= at and _position <= at+elem.length:
            return i
    raise ValueError("No element found at position: {0}".format(position))


def get_element_count(sequence_name: str) -> int:
    """
    Return number of elements in the original sequence.

    :param str sequence_name: sequence name
    :returns: number of elements in the original sequence
    :raises ValueError: if the sequence is invalid.
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.nodes.curr


def get_expanded_element(sequence_name: str, element_index: int) -> dict:
    """
    Return requested element in the expanded sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the element with the specified index
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range

    NOTE: this function may currently return elements beyond the end of the
    expanded sequence if requested to do so.
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index > seq.n_nodes:
        raise IndexError(
            "Index out of range: {0} (element count is {1})"
            .format(element_index, seq.n_nodes))
    return _get_node(seq.all_nodes[element_index], seq.ref_flag, seq.n_nodes > 0, seq.line)


def get_expanded_element_positions(sequence_name: str) -> list:
    """
    Get list with positions of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :returns: positions of all elements in the sequence
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef int i
    cdef clib.node** nodes = seq.all_nodes
    return [_get_node_entry_pos(nodes[i], seq.ref_flag, seq.n_nodes > 0)
            for i in range(seq.n_nodes)]


def get_expanded_element_names(sequence_name: str) -> list:
    """
    Get list with the names of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :returns: names of all elements in the sequence
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef int i
    return [_node_name(seq.all_nodes[i])
            for i in range(seq.n_nodes)]


def get_expanded_element_name(sequence_name: str, element_index: int) -> str:
    """
    Get list with the names of all elements of a specific sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the name of the element with the specified index
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index > seq.n_nodes:
        raise IndexError(
            "Index out of range: {0} (element count is {1})"
            .format(element_index, seq.n_nodes))
    return _node_name(seq.all_nodes[element_index])


def get_expanded_element_index(sequence_name: str, element_name: str) -> int:
    """
    Return index of element with specified name in the expanded sequence.

    NOTE: this is the brute-force linear-time algorithm and therefore not
    recommended for frequent execution.

    :param str sequence_name: sequence name
    :param str element_name: element index
    :returns: the index of the specified element, -1 if not found
    :raises ValueError: if the sequence is invalid
    """
    # Unfortunately, there is no name_list for the expanded list. Therefore,
    # Therefore, we can only provide a linear-time lookup.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef bytes _element_name = _cstr(name_to_internal(element_name))
    for i in range(seq.n_nodes):
        if seq.all_nodes[i].name == _element_name:
            return i
    raise ValueError("Element name not found: {0!r}".format(element_name))


def get_expanded_element_index_by_position(sequence_name: str, position: float) -> int:
    """
    Return index of element at specified position in the expanded sequence.

    :param str sequence_name: sequence name
    :param float position: position (S coordinate)
    :returns: the index of an element at that position
    :raises ValueError: if the sequence or element name is invalid
    """
    # This is implemented here just for performance, but still uses a linear
    # algorithm. If you want more, you need to copy the elements over and use
    # a suitable ordered data type.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef double _position = position
    cdef clib.node* elem
    cdef double at
    for i in range(seq.n_nodes):
        elem = seq.all_nodes[i]
        at = _get_node_entry_pos(elem, seq.ref_flag, seq.n_nodes > 0)
        if _position >= at and _position <= at+elem.length:
            return i
    raise ValueError("No element found at position: {0}".format(position))


def get_expanded_element_count(sequence_name: str) -> int:
    """
    Return number of elements in the expanded sequence.

    :param str sequence_name: sequence name
    :returns: number of elements in the expanded sequence
    :raises ValueError: if the sequence is invalid.
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.n_nodes


def get_global_element(element_index: int) -> dict:
    """
    Return requested element in the expanded sequence.

    :param int element_index: element index
    :returns: the element with the specified index
    :raises IndexError: if the index is out of range
    """
    cdef clib.el_list* elems = clib.element_list
    if element_index < 0 or element_index >= elems.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, elems.curr))
    return _get_element(elems.elem[element_index])


def get_global_element_name(element_index: int) -> str:
    """
    Return name of element.

    :param int element_index: element index
    :returns: element name
    :raises IndexError: if the index is out of range
    """
    cdef clib.el_list* elems = clib.element_list
    if element_index < 0 or element_index >= elems.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, elems.curr))
    return _str(elems.list.names[element_index])


def get_global_element_index(element_name: str) -> int:
    """
    Return index of element with specified name in the global element list.

    :param str element_name: element index
    :returns: the index of the specified element, -1 if not found
    """
    cdef bytes _element_name = _cstr(element_name)
    return clib.name_list_pos(_element_name, clib.element_list.list)


def get_global_element_count() -> int:
    """
    Return number of globally visible elements.
    """
    return clib.element_list.curr


def get_base_type_names() -> list:
    """
    Return list of element names for base types.
    """
    return _name_list(clib.base_type_list.list)


def get_defined_command(command_name: str) -> dict:
    """Return MAD-X command as dict of values."""
    cdef bytes _command_name = _cstr(command_name)
    cdef int index = clib.name_list_pos(_command_name, clib.defined_commands.list)
    if index == -1:
        raise ValueError("Invalid command: {!r}".format(command_name))
    return _parse_command(clib.defined_commands.commands[index])


def get_defined_command_names() -> list:
    """Return list of MAD-X command names."""
    return _name_list(clib.defined_commands.list)


def is_sequence_expanded(sequence_name: str) -> bool:
    """
    Check whether a sequence has already been expanded.

    :param str sequence_name: sequence name
    :returns: expanded state of the sequence
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.n_nodes > 0


def eval(expression: str) -> float:
    """
    Evaluates an expression and returns the result as double.

    :param str expression: symbolic expression to evaluate
    :returns: numeric value of the expression

    NOTE: This function does not perform rigorous input validation! It uses
    nothing but the MAD-X builtin rather incompetent error checks. This means
    invalid input such as '+' can lead to program crashes! If you're looking
    for more secure validation, see :func:`cpymad.util.check_expression`.
    """
    cdef char* string = _polish_expr(expression)
    return clib.polish_value(clib.deco, string)


def expression_vars(expression: str) -> set:
    """
    Returns all the variables in an expression.

    :param str expression: symbolic expression
    :returns: set of variables

    NOTE: This function does not perform rigorous input validation! It uses
    nothing but the MAD-X builtin rather incompetent error checks. This means
    invalid input such as '+' can lead to program crashes! If you're looking
    for more secure validation, see :func:`cpymad.util.check_expression`.
    """
    _polish_expr(expression)
    cdef int i
    vars = {_str(clib.expr_chunks.names[k % 100000000])
            for i in range(clib.deco.curr)
            for k in [clib.deco.i[i]]
            if k // 100000000 == 1}
    return vars


# Helper functions:

# The following functions are `cdef functions`, i.e. they can only be
# called from Cython code. It is necessary to use `cdef functions` whenever
# we want to pass parameters or return values with a pure C type.


cdef char* _polish_expr(str expression) except NULL:
    # TODO: This function uses global variables as temporaries - which is in
    # general an *extremely* bad design choice. Even though MAD-X uses global
    # variables internally anyway, this is no excuse for cpymad to copy that
    # behaviour.
    # TODO: not sure about the flags (the magic constants 0, 2)
    cdef bytes _expr = _cstr(expression.lower())
    clib.pre_split(_expr, clib.c_dum, 0)
    clib.check_table(clib.c_dum.c)
    clib.check_tabindex(clib.c_dum.c)
    clib.check_tabstring(clib.c_dum.c)
    clib.mysplit(clib.c_dum.c, clib.tmp_p_array)
    # NOTE: `loc_expr` is mostly useless for input validation. It even accepts
    # stuff such as '+' that will lead to program crashes on evaluation. We
    # use it nevertheless to get a minimal amount of error-checking. Not using
    # cpymad.util.check_expression for simplicity and performance.
    cdef int start=0, stop
    etype = clib.loc_expr(clib.tmp_p_array.p,
                          clib.tmp_p_array.curr,
                          start, &stop)
    if etype == 0 or stop+1 < clib.tmp_p_array.curr:
        raise ValueError("Invalid expression: {!r}".format(expression))
    if clib.polish_expr(stop + 1, clib.tmp_p_array.p) != 0:
        raise ValueError("Invalid expression: {!r}".format(expression))
    return clib.join(clib.tmp_p_array.p, stop + 1)


_expr_types = [bool, int, float]

cdef _expr(clib.expression* expr,
           double value,
           int typeid=clib.PARAM_TYPE_DOUBLE):
    """Return a parameter value with an appropriate type."""
    _type = _expr_types[typeid]
    _value = _type(value)
    if expr is NULL or expr.string is NULL:
        return _value, None
    _expr = _str(expr.string).strip()
    if not _expr:
        return _value, None
    try:
        return _type(float(_expr)), None
    except ValueError:
        pass
    value = clib.expression_value(expr, 2)
    return _type(value), _expr


cdef _get_param_value(clib.command_parameter* par):

    """
    Get the value of a command parameter.

    :returns: value of the parameter
    :rtype: depends on the parameter type
    :raises ValueError: if the parameter type is invalid
    """

    if par.type in (clib.PARAM_TYPE_LOGICAL,
                    clib.PARAM_TYPE_INTEGER,
                    clib.PARAM_TYPE_DOUBLE):
        return _expr(par.expr, par.double_value, par.type)

    if par.type == clib.PARAM_TYPE_STRING:
        return _str(par.string), None

    if par.type == clib.PARAM_TYPE_CONSTRAINT:
        val = min = max = (None, None)
        if par.c_type == clib.CONSTR_TYPE_NONE: # occurs in defined_commands
            pass
        if par.c_type == clib.CONSTR_TYPE_MIN:
            min = _expr(par.min_expr, par.c_min)
        if par.c_type == clib.CONSTR_TYPE_MAX:
            max = _expr(par.max_expr, par.c_max)
        if par.c_type == clib.CONSTR_TYPE_BOTH:
            min = _expr(par.min_expr, par.c_min)
            max = _expr(par.max_expr, par.c_max)
        if par.c_type == clib.CONSTR_TYPE_VALUE:
            val = _expr(par.expr, par.double_value)
        return (Constraint(val=val[0], min=min[0], max=max[0]),
                Constraint(val=val[1], min=min[1], max=max[1]))

    if par.type in (clib.PARAM_TYPE_INTEGER_ARRAY, clib.PARAM_TYPE_DOUBLE_ARRAY):
        fields = [
            _expr(NULL if par.expr_list is NULL else par.expr_list.list[i],
                  par.double_array.a[i],
                  par.type - clib.PARAM_TYPE_LOGICAL_ARRAY)
            for i in range(par.double_array.curr)
        ]
        return ([f[0] for f in fields],
                [f[1] for f in fields])

    if par.type == clib.PARAM_TYPE_STRING_ARRAY:
        return [_str(par.m_string.p[i])
                for i in range(par.m_string.curr)], None

    raise ValueError("Unknown parameter type: {}".format(par.type))


cdef dict _parse_command(clib.command* cmd):
    """
    Get the values of all parameters of a command.

    :returns: the command parameters
    """
    cdef int i
    return {'name': _str(cmd.name), 'data': {
        # enforce lower-case keys:
        name: Parameter(
            name,
            *_get_param_value(cmd.par.parameters[i]),
            dtype=cmd.par.parameters[i].type,
            inform=cmd.par_names.inform[i])
        for i in range(cmd.par.curr)
        for name in [_str(cmd.par.parameters[i].name).lower()]
    }}


# The 'except NULL' clause is needed to forward exceptions from cdef
# functions with C return values, see:
# http://docs.cython.org/src/userguide/language_basics.html#error-return-values
cdef clib.sequence* _find_sequence(str sequence_name) except NULL:
    """
    Get pointer to the C sequence struct of the specified sequence or NULL.

    :param str sequence_name: sequence name
    :raises ValueError: if the sequence can not be found
    """
    cdef bytes _sequence_name = _cstr(sequence_name)
    cdef clib.sequence_list* seqs = clib.madextern_get_sequence_list()
    cdef int index = clib.name_list_pos(_sequence_name, seqs.list)
    if index == -1:
        raise ValueError("Invalid sequence: {!r}".format(sequence_name))
    return seqs.sequs[index]


cdef clib.table* _find_table(str table_name) except NULL:
    cdef bytes _table_name = _cstr(table_name)
    cdef int index = clib.name_list_pos(_table_name, clib.table_register.names)
    if index == -1:
        raise ValueError("Invalid table: {!r}".format(table_name))
    return clib.table_register.tables[index]


cdef _split_header_line(header_line):
    """Parse a table header value."""
    _, key, kind, value = _str(header_line).split(None, 3)
    key = key.lower()
    if kind == "%le":
        return key, float(value)    # convert to number
    elif kind.endswith('s'):
        return key, value[1:-1]     # strip quotes from string
    else:
        return key, value           #


cdef _name_list(clib.name_list* names):
    """Return a python list of names for the name_list."""
    cdef int i
    return [_str(names.names[i]) for i in range(names.curr)]


# Leave return type unspecified to avoid cython error with unicode.split(None):
# ("TypeError: coercing to Unicode: need string or buffer, NoneType found")
cdef _str(const char* s):
    """Decode C string to python string."""
    if s is NULL:
        return ""
    return s.decode('utf-8')


cdef bytes _cstr(s):
    """Encode python string to C string."""
    if s is None:
        return b""
    if isinstance(s, bytes):
        return <bytes> s
    return <bytes> s.encode('utf-8')


cdef str _node_name(clib.node* node):
    return name_from_internal(_str(node.name))


cdef dict _get_node(clib.node* node, int ref_flag, int is_expanded, int line):
    """Return dictionary with node + element attributes."""
    if node.p_elem is not NULL:
        data = _get_element(node.p_elem)
    elif node.p_sequ is not NULL:
        data = {
            'name': _str(node.p_sequ.name),
            'base_type': 'sequence',
            'data': {},
        }
    else:
        # Maybe this is a valid case, but better detect it with boom!
        raise RuntimeError(
            "Empty node or subsequence! Please report this incident!")

    if not line:
        # Update `at` command parameters in order to avoid surprises when
        # specifying `at` value for elements defined outside the sequence:
        data['data']['at'] = Parameter(
            'at', *_expr(node.at_expr, node.at_value),
            dtype=clib.PARAM_TYPE_DOUBLE, inform=1)
        data['data']['from'] = Parameter(
            'from', _str(node.from_name), None,
            dtype=clib.PARAM_TYPE_STRING, inform=node.from_name != NULL)
    data['node_name'] = _node_name(node)
    data['occ_cnt'] = node.occ_cnt
    data['enable'] = node.enable
    data['base_name'] = _str(node.base_name)
    data['position'] = _get_node_entry_pos(node, ref_flag, is_expanded)
    data['length'] = node.length
    data['align_errors'] = None if node.p_al_err is NULL else AlignError(
        *_memview(node.p_al_err))
    data['field_errors'] = None if node.p_fd_err is NULL else FieldError(
        dkn=list(_memview(node.p_fd_err)[0::2]),
        dks=list(_memview(node.p_fd_err)[1::2]))
    data['phase_errors'] = None if node.p_ph_err is NULL else PhaseError(
        dpn=list(_memview(node.p_ph_err)[0::2]),
        dps=list(_memview(node.p_ph_err)[1::2]))
    data['chkick'] = node.chkick
    data['cvkick'] = node.cvkick
    return data


cdef double [:] _memview(clib.double_array* array):
    return <double [:array.curr]> array.a


cdef double _get_node_entry_pos(clib.node* node, int ref_flag, int is_expanded):
    """Normalize 'at' value to node entry."""
    if is_expanded:
        return node.position - node.length / 2
    cdef double position = node.at_value
    if ref_flag == clib.REF_CENTER:
        return position - node.length / 2
    elif ref_flag == clib.REF_EXIT:
        return position - node.length
    else:
        return position


cdef dict _get_element(clib.element* elem):
    """Return dictionary with element attributes."""
    data = _parse_command(elem.def_)
    data['name'] = _str(elem.name)
    data['length'] = elem.length
    data['parent'] = _str(elem.parent.name)
    data['base_type'] = _str(elem.base_type.name)
    return data


cdef str _get_table_row_name(clib.table* table, int index):
    if index < 0 or index >= table.curr:
        raise ValueError("Invalid row index: {}".format(index))
    return normalize_range_name(name_from_internal(_str(
        table.node_nm.p[index]
    )))


cdef clib.variable* _get_var(name) except NULL:
    cdef bytes _name = _cstr(name.lower())
    cdef clib.variable* var = clib.find_variable(_name, clib.variable_list)
    if var is NULL:
        raise KeyError("Variable not defined: {!r}".format(name))
    return var


cdef void _strip_comments(char* text) nogil:
    cdef char* dest = text
    cdef char c, d
    while True:
        c = text[0]
        if c == 0:
            break
        d = text[1]
        if c == b'!' or (c == b'/' and d == b'/'):
            text = strchr(text+1, b'\n')
            if text == NULL:
                break
            continue
        if c == b'/' and d == b'*':
            text = strstr(text+2, b"*/")
            if text == NULL:
                break
            text += 2
            continue
        dest[0] = text[0]
        text += 1
        dest += 1
    dest[0] = 0



"""
Tests for the functionality in :mod:`cpymad.util`.
"""

from cpymad import util
from cpymad.madx import Madx, AttrDict
from cpymad.types import Range, Constraint
import pytest


@pytest.fixture
def mad():
    with Madx(prompt='X:> ') as mad:
        yield mad


def is_valid_expression(expr):
    try:
        return util.check_expression(expr)
    except ValueError:
        return False


def test_mad_quote():
    # default to double quotes:
    assert util.mad_quote('foo bar') == '"foo bar"'
    # fallback to single quotes:
    assert util.mad_quote('"foo bar"') == '\'"foo bar"\''
    # MAD-X doesn't know escapes:
    with pytest.raises(ValueError):
        util.mad_quote('"foo" \'bar\'')


def test_is_identifier():
    assert util.is_identifier('hdr23_oei')
    assert util.is_identifier('_hdr23_oei')
    assert not util.is_identifier('2_hdr23_oei')
    assert not util.is_identifier('hdr oei')
    assert not util.is_identifier('hdr@oei')


def test_expr_symbols():
    assert util.expr_symbols('foobar') == {'foobar'}
    assert util.expr_symbols('foo*bar') == {'foo', 'bar'}
    assert util.expr_symbols('quad->k1') == {'quad->k1'}
    assert util.expr_symbols('q->k-p->k') == {'q->k', 'p->k'}
    assert util.expr_symbols('a * sin(x)') == {'a', 'sin', 'x'}


def test_name_from_internal():
    assert util.name_from_internal('foo.23:1') == 'foo.23'
    assert util.name_from_internal('foo.23:43') == 'foo.23[43]'
    with pytest.raises(ValueError):
        util.name_from_internal('foo:23o')


def test_name_to_internal():
    assert util.name_to_internal('foo.23') == 'foo.23:1'
    assert util.name_to_internal('foo.23[43]') == 'foo.23:43'
    with pytest.raises(ValueError):
        util.name_to_internal('foo:23o')


def test_normalize_range_name():
    assert util.normalize_range_name('dr[1]') == 'dr[1]'
    assert util.normalize_range_name('lebt$start') == '#s'
    assert util.normalize_range_name('lebt$end') == '#e'
    assert util.normalize_range_name('dr') == 'dr'
    assert util.normalize_range_name('l$start/qp[5]') == '#s/qp[5]'


def test_format_command(mad):
    assert util.format_command(
        'twiss', sequence='lhc'
    ) == 'twiss, sequence="lhc";'
    assert util.format_command(
        'option', echo=True
    ) == 'option, echo=true;'
    assert util.format_command(
        'constraint', betx=Constraint(max=3.13)
    ) == 'constraint, betx<3.13;'
    assert util.format_command(
        mad.command.quadrupole, k1='pi/2'
    ) == 'quadrupole, k1:=pi/2;'
    assert util.format_command(
        'multipole', knl=[0.0, 1.0, 2.0]
    ) == 'multipole, knl={0.0,1.0,2.0};'
    assert util.format_command(
        'twiss', range=Range('#s', '#e')
    ) == 'twiss, range=#s/#e;'
    assert util.format_command(
        'select', class_='quadrupole',
    ) == 'select, class="quadrupole";'

    assert util.format_command(
        mad.elements.quadrupole, k1="hello + world"
    ) == 'quadrupole, k1:=hello + world;'
    assert util.format_command(
        # match->sequence parameter is list in MAD-X!
        mad.command.match, sequence="foo"
    ) == "match, sequence=foo;"
    assert util.format_command(
        mad.command.select, class_='quadrupole',
    ) == 'select, class=quadrupole;'


def test_format_param():
    fmt = util.format_param

    assert fmt('foo', None) is None

    assert fmt('foo', Range('#s', '#e')) == 'foo=#s/#e'

    assert fmt('foo', Constraint(min="bar")) == 'foo>bar'
    assert fmt('foo', Constraint(max="bar")) == 'foo<bar'
    assert fmt('foo', Constraint("bar"))     == 'foo=bar'
    assert fmt('foo', Constraint("bar", min=-2, max=3)) == 'foo>-2, foo<3'

    assert fmt('foo', True) == 'foo=true'
    assert fmt('foo', False) == 'foo=false'

    assert fmt('range', 'l$start/l$end') == 'range=#s/#e'
    assert fmt('range', 'qp[5]') == 'range=qp[5]'
    assert fmt('range', 'qp[5]') == 'range=qp[5]'
    assert fmt('range', ('l$start', 'qp[1]')) == 'range=#s/qp[1]'

    assert fmt('file', "Weird N\ame!") == 'file="Weird N\ame!"'
    assert fmt('fool', "Weird N\ame!") == 'fool="weird n\ame!"'

    assert fmt('knl', [1.5, 'k1', 'k2']) == 'knl={1.5,k1,k2}'

    assert fmt('pi', 3) == 'pi=3'
    assert fmt('pi', 3.14) == 'pi=3.14'


def test_format_cmdpar(mad):
    fmt = util.format_cmdpar

    mult = mad.elements.quadrupole
    cons = mad.command.constraint
    match = mad.command.match
    twiss = mad.command.twiss

    assert fmt(mult, 'knl', None) == ''

    assert fmt(twiss, 'range', 'h$start/h$end') == 'range=#s/#e'
    assert fmt(twiss, 'range', ('a', 'b')) == 'range=a/b'
    assert fmt(match, 'range', '#s/#e') == 'range=#s/#e'
    assert fmt(match, 'range', ['#s/#e', 's/e']) == 'range={#s/#e,s/e}'

    assert fmt(cons, 'betx', Constraint(min="bar")) == 'betx>bar'
    assert fmt(cons, 'betx', Constraint(max="bar")) == 'betx<bar'
    assert fmt(cons, 'betx', Constraint(max="bar")) == 'betx<bar'
    assert fmt(
        cons, 'betx', Constraint("bar", min=-2, max=3)
    ) == 'betx>-2, betx<3'
    assert fmt(cons, 'betx', Constraint("bar")) == 'betx=bar'
    assert fmt(cons, 'betx', "bar")             == 'betx:=bar'
    assert fmt(cons, 'betx', 3.14)              == 'betx=3.14'

    assert fmt(match, 'slow', True) == 'slow=true'
    assert fmt(match, 'slow', False) == 'slow=false'

    assert fmt(twiss, 'file', "Weird N\ame!") == 'file="Weird N\ame!"'

    assert fmt(mult, 'knl', [1.5, 'x1', 'x2']) == 'knl:={1.5,x1,x2}'
    assert fmt(mult, 'knl', [1.5, 2.5, 3.5]) == 'knl={1.5,2.5,3.5}'

    assert fmt(mult, 'kmin', 3) == 'kmin=3'
    assert fmt(mult, 'kmin', 3.14) == 'kmin=3.14'
    assert fmt(mult, 'kmin', "x/(b+c)") == 'kmin:=x/(b+c)'

    assert fmt(match, 'sequence', 'seq') == 'sequence=seq'
    assert fmt(match, 'sequence', ['s0', 's1']) == 'sequence={s0,s1}'

    # missing attributes:
    with pytest.raises(KeyError):
        fmt(mult, 'foo', None)


def test_check_expression():
    assert is_valid_expression('a*b')
    assert is_valid_expression('a * (qp->k1+7.) ^ .5e-3')
    assert is_valid_expression('-a*-3')
    assert is_valid_expression('E-m*c^2')
    assert is_valid_expression('-(x / y)')
    assert is_valid_expression('1 ^(2)')
    assert is_valid_expression('+ 1')
    assert is_valid_expression('x')
    assert is_valid_expression('1')
    assert is_valid_expression('1 + 2')
    assert is_valid_expression('sin(x)')
    assert is_valid_expression('1 + table(twiss, mqf, x)')

    assert not is_valid_expression('')
    assert not is_valid_expression('x x')
    assert not is_valid_expression('+')
    assert not is_valid_expression('*')
    assert not is_valid_expression('* 1')
    assert not is_valid_expression('1 +')
    assert not is_valid_expression('sin(x) +')
    assert not is_valid_expression('sin(x) + * ')

    assert not is_valid_expression('E=m*c^2')
    assert not is_valid_expression('(@)')
    assert not is_valid_expression('(()')
    assert not is_valid_expression('(+3')
    assert not is_valid_expression('1*')     # NOTE: valid in MAD-X
    assert not is_valid_expression('3+')     # NOTE: valid in MAD-X
    assert not is_valid_expression('()')
    assert not is_valid_expression('(1 | 2)')

    assert not is_valid_expression('1 2')
    assert not is_valid_expression('1 (2)')
    assert not is_valid_expression('^(2)')


def test_attrdict():
    pi = 3.14
    d = AttrDict({'foo': 'bar', 'pi': pi})
    assert d.foo == 'bar'
    assert d.pi == pi
    assert d['foo'] == 'bar'
    assert d['pi'] == pi
    assert len(d) == 2
    assert set(d) == {'foo', 'pi'}
    assert 'FoO' in d
    assert 'pI' in d
    assert 'foopi' not in d
    assert 'foo' in str(d)
    assert "'bar'" in str(d)
    assert 'pi' in str(d)
    assert '3.14' in str(d)
    with pytest.raises(AttributeError):
        d.foobar
    with pytest.raises(KeyError):
        d['foobar']
    with pytest.raises(AttributeError):
        del d.foo


# TODO: test other functions



# encoding: utf-8
"""
Setup script for cpymad.

This script is meant only for packagers and developers and can be used to
install cpymad or create installers assuming you have already built MAD-X
beforehand.

For more information, see
    http://hibtc.github.io/cpymad/installation
"""

# Now that we distribute wheels for most supported platforms we do not attempt
# to build MAD-X automatically! These are the preferred method of installation
# for most users while for packagers/developers it is preferrable to have
# direct control over the MAD-X build process itself. Therefore, there are few
# who would actually profit from automatic setup. Furthermore, building MAD-X
# in setup.py adds a lot of complexity. If you don't believe me, take a look
# at the commit that first added this paragraph (can be identified using `git
# blame`) and the simplifications that were possible in the following commits.

from setuptools import setup, find_packages, Extension
from distutils.util import get_platform
from distutils import sysconfig
from argparse import ArgumentParser

import sys
import os

try:
    # Use Cython if available:
    from Cython.Build import cythonize
except ImportError:
    # Otherwise, use the shipped .c file:
    def cythonize(extensions):
        return extensions

# Windows:  win32/win-amd64
# Linux:    linux-x86_64/...
# Mac:      darwin*
IS_WIN = get_platform().startswith('win')
IS_ARM = get_platform().startswith('linux-aarch')


# We parse command line options using our own mechanim. We could use
# build_ext.user_options instead, but then these parameters can be passed
# only to the 'build_ext' command, not to 'build', 'develop', or
# 'install'.
def command_line_options():
    usage = 'setup.py <command> [options]'
    parser = ArgumentParser(description=__doc__, usage=usage)
    parser.add_argument(
        '--madxdir', dest='madxdir',
        default=os.environ.get('MADXDIR'),
        help='MAD-X installation prefix')
    option(parser, 'static', 'do {NOT}use static linkage')
    option(parser, 'shared', 'MAD-X was {NOT}built with BUILD_SHARED_LIBS')
    option(parser, 'lapack', 'MAD-X was {NOT}built with LAPACK')
    option(parser, 'blas', 'MAD-X was {NOT}built with BLAS')
    option(parser, 'X11', 'MAD-X was {NOT}built with MADX_X11')
    option(parser, 'quadmath', 'do {NOT}link against libquadmath')
    return parser


def option(parser, name, descr):
    """Add a negatable option to parser."""
    env_var = os.environ.get(name.upper())
    default = bool(int(env_var)) if env_var else None
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--' + name, dest=name,
        default=default,
        action='store_true', help=descr.format(NOT=''))
    group.add_argument(
        '--no-' + name, dest=name,
        action='store_false', help=descr.format(NOT='not '))
    return group


def fix_distutils_sysconfig_mingw():
    """
    When using windows and MinGW, in distutils.sysconfig the compiler (CC) is
    not initialized at all, see http://bugs.python.org/issue2437. The
    following manual fix for this problem may cause other issues, but it's a
    good shot.
    """
    if sysconfig.get_config_var('CC') is None:
        sysconfig._config_vars['CC'] = 'gcc'


def exec_file(path):
    """Execute a python file and return the `globals` dictionary."""
    namespace = {}
    with open(path, 'rb') as f:
        exec(f.read(), namespace, namespace)
    return namespace


def get_extension_args(madxdir, shared, static, **libs):
    """Get arguments for C-extension (include pathes, libraries, etc)."""
    # libquadmath isn't available on aarch64, see #101:
    if libs.get('quadmath') is None:
        libs['quadmath'] = not IS_ARM
    include_dirs = []
    library_dirs = []

    if madxdir:
        prefix = os.path.expanduser(madxdir)
        include_dirs += [os.path.join(prefix, 'include')]
        library_dirs += [os.path.join(prefix, 'lib'),
                         os.path.join(prefix, 'lib64')]

    libraries = ['madx']
    if not shared:
        # NOTE: If MAD-X was built with BLAS/LAPACK, you must manually provide
        # arguments `python setup.py build_ext -lblas -llapack`!
        libraries += ['DISTlib', 'ptc', 'gc-lib', 'stdc++', 'gfortran']
        libraries += [lib for lib, use in libs.items() if use]

    link_args = ['--static'] if static and not IS_WIN else []

    return dict(
        libraries=libraries,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        runtime_library_dirs=library_dirs if shared else [],
        extra_compile_args=['-std=gnu99'],
        extra_link_args=link_args,
    )


if __name__ == '__main__':
    sys.path.append(os.path.dirname(__file__))
    fix_distutils_sysconfig_mingw()
    options, sys.argv[1:] = command_line_options().parse_known_args()
    metadata = exec_file('src/cpymad/__init__.py')
    setup(
        name='cpymad',
        version=metadata['__version__'],
        description=metadata['__summary__'],
        ext_modules=cythonize([
            Extension('cpymad.libmadx',
                      sources=["src/cpymad/libmadx.pyx"],
                      **get_extension_args(**options.__dict__)),
        ]),
        packages=find_packages('src'),
        package_dir={'': 'src'},
        zip_safe=False,             # zip is bad for redistributing shared libs
        include_package_data=True,  # include files matched by MANIFEST.in
        install_requires=[
            'importlib_resources',
            'numpy',
            'minrpc>=0.0.8',
        ],
    )



from __future__ import unicode_literals
import sys
import warnings


__title__ = 'cpymad'
__version__ = '1.10.0'

__summary__ = 'Cython binding to MAD-X'
__uri__ = 'https://github.com/hibtc/cpymad'

__credits__ = """
Current cpymad maintainer:

    - Thomas Gläßle <t_glaessle@gmx.de>

Initial pymad creators:

    - Yngve Inntjore Levinsen <Yngve.Inntjore.Levinsen@cern.ch>
    - Kajetan Fuchsberger <Kajetan.Fuchsberger@cern.ch>
"""


if sys.version_info < (3, 6):
    _unsupported_version = (
        "Support for python 3.5 and below will be removed in a future release!\n"
        "If you need continued support for an older version, let us know at:\n"
        "  https://github.com/hibtc/cpymad/issues")
    warnings.warn(_unsupported_version, DeprecationWarning)


def get_copyright_notice() -> str:
    from importlib_resources import read_text
    return read_text('cpymad.COPYING', 'cpymad.rst', encoding='utf-8')
